
-- # import PropertyUP

-- Categories -------------------------------------------------------------

IsCategory : (Object : Type) -> (Arrow : Object -> Object -> Type) ->
	      ((A : Object) -> Arrow A A) ->
	      ((A, B, C : Object) -> Arrow B C -> Arrow A B -> Arrow A C) ->
	      Type
IsCategory Object Arrow id compose =
	      { !leftId :  (A, B : Object) ->
			    (f : Arrow A B) ->
			    Eq (Arrow A B)
			       (compose A B B (id B) f)
			       f
	      ; !rightId : (A, B : Object) ->
			    (f : Arrow A B) ->
			    Eq (Arrow A B)
			       (compose A A B f (id A))
			       f
	      ; !assoc :   (A, B, C, D : Object) ->
			    (f : Arrow C D) ->
			    (g : Arrow B C) ->
			    (h : Arrow A B) ->
			    Eq (Arrow A D)
			       (compose A B D
				 (compose B C D f g)
				 h
			       )
			       (compose A C D
				 f
				 (compose A B C g h)
			       )
	      }

Category : Type
Category =  { !Object	: Type
	    ; !Arrow	: Object -> Object -> Type
	    ; !id	: (A:Object) -> Arrow A A
	    ; !compose	: (A,B,C:Object) ->
			   Arrow B C -> Arrow A B -> Arrow A C
	    ; !proofs	: IsCategory Object Arrow id compose
	    }

-- Isomorphic objects -----------------------------------------------------

isomorphic : (CC : Category)	  ->
	      (A, B : CC.Object) ->
	      Type
isomorphic CC A B =
    { !i  : CC.Arrow A B
    ; !j  : CC.Arrow B A
    ; !ij : Eq (CC.Arrow B B) (CC.compose B A B i j) (CC.id B)
    ; !ji : Eq (CC.Arrow A A) (CC.compose A B A j i) (CC.id A)
    }

-- Unique arrows ----------------------------------------------------------

UniqueArrow : (CC : Category) -> (A, B : CC.Object) -> Type
UniqueArrow CC A B =
    { !arrow  : CC.Arrow A B
    ; !unique : (h : CC.Arrow A B) -> Eq (CC.Arrow A B) h arrow
    }

-- Dual categories --------------------------------------------------------

op : Category -> Category
op CC =  ( CC.Object
	  , \A B -> CC.Arrow B A
	  , CC.id
	  , \A B C f g -> CC.compose C B A g f
	  , ( \A B f	      -> fol-plugin(CC.proofs.rightId B A f)
	    , \A B f	      -> fol-plugin(CC.proofs.leftId B A f)
	    , \A B C D f g h  -> fol-plugin(CC.proofs.assoc D C B A h g f)
	    )
	  )

-- Initial and terminal objects -------------------------------------------

IsTerminal : (CC : Category) -> CC.Object -> Type
IsTerminal CC One = (A : CC.Object) -> UniqueArrow CC A One

TerminalObject : (CC : Category) -> Type
TerminalObject CC =
    { !One  : CC.Object
    ; !bang : IsTerminal CC One
    }

IsInitial : (CC : Category) -> CC.Object -> Type
IsInitial CC Zero = (A : CC.Object) -> UniqueArrow CC Zero A

InitialObject : (CC : Category) -> Type
InitialObject CC =
    { !Zero : CC.Object
    ; !bang : IsInitial CC Zero
    }

initialTerminalDual : (CC : Category) ->
		       (T : TerminalObject CC) ->
		       IsInitial (op CC) T.One
initialTerminalDual CC T = T.bang

terminalInitialDual : (CC : Category) ->
		       (I : InitialObject CC) ->
		       IsTerminal (op CC) I.Zero
terminalInitialDual CC I = I.bang

terminalObjectIsUnique :
    (CC : Category) ->
    (T1, T2 : TerminalObject CC) ->
    isomorphic CC T1.One T2.One
terminalObjectIsUnique CC T1 T2 =
    ( (T2.bang T1.One).arrow
    , (T1.bang T2.One).arrow
    , fol-plugin((T2.bang T2.One).unique)
    , fol-plugin((T1.bang T1.One).unique)
    )

initialObjectIsUnique :
    (CC : Category) ->
    (I1, I2 : InitialObject CC) ->
    isomorphic CC I1.Zero I2.Zero
initialObjectIsUnique CC I1 I2 = (p.j, p.i, p.ij, p.ji)
  where
    p = terminalObjectIsUnique (op CC) (I1.Zero,I1.bang) (I2.Zero,I2.bang)

