-- Insertion sort example for the quickCheck plugin.
-- Author:: Ulf Norell
-- Created:: 2004-12-09

-- # import QuickCheck
-- # import Bool.Proofs

-- Insertion sort ---------------------------------------------------------

Leq : (A : Set) -> Set
Leq A = A -> A -> Bool

ins : (A : Set) -> Leq A -> A -> List A -> List A
ins A _ x [] = [x]
ins A (<=) x (y::ys) =
    if (List A) (x <= y)
	(x :: y :: ys)
	(y :: ins A (<=) x ys)

isort : (A : Set) -> Leq A -> List A -> List A
isort A (<=) [] = []
isort A (<=) (x::xs) = ins A (<=) x (isort A (<=) xs)

-- Testing ----------------------------------------------------------------

-- Helpers
ordered : List Nat -> Bool
ordered []	  = true
ordered [x]	  = true
ordered (x::y::xs)  = x <= y && ordered (y::xs)

-- Generators


-- Ordered lists
orderedGen : Gen (List Nat)
orderedGen = liftM  Gen monadGen
		    (List Nat) (List Nat)
		    (scanl1 Nat (+)) (arbList Nat arbNat)

-- Properties
PropInsertBad = (x:Nat) -> (xs:List Nat) ->
		    Holds (ordered xs) ==> 
		      collect Nat (length Nat xs)
			(Holds (ordered (ins Nat (<=) x xs)))

test_insBad : PropInsertBad
test_insBad = qc-plugin(arbNat, arbList Nat arbNat)

-- We don't have to use forAll here, but since we are relying on the
-- fact that the list is ordered it's good to specify the generator
-- in the property and not leave it up to the tester.
PropInsertGood = (x:Nat) ->
		    forAll (List Nat) orderedGen (\xs ->
		      collect Nat (length Nat xs)
			(Holds (ordered (ins Nat (<=) x xs)))
		    )

test_insGood : PropInsertGood
test_insGood = qc-plugin(arbNat)

PropInsertLength = (x : Nat) -> (xs : List Nat) ->
		      Holds (succ (length Nat xs) ==
			     length Nat (ins Nat (<=) x xs))

test_insLength : PropInsertLength
test_insLength = qc-plugin(arbNat, arbList Nat arbNat)

